@@ -16,10 +16,8 @@ pub fn target() -> Target {
         executables: true,
         relocation_model: RelocModel::Static,
         disable_redzone: true,
-        linker_is_gnu: true,
         max_atomic_width: Some(128),
         panic_strategy: PanicStrategy::Abort,
-        unsupported_abis: super::arm_base::unsupported_abis(),
         ..Default::default()
     };
     Target {
