@@ -1,18 +1,11 @@
 use std::env;
 use std::fs;
-// ANCHOR: here
-use std::process;
 
+// ANCHOR: here
 fn main() {
     let args: Vec<String> = env::args().collect();
 
-    let config = Config::new(&args).unwrap_or_else(|err| {
-        println!("Problem parsing arguments: {}", err);
-        process::exit(1);
-    });
-
-    // --snip--
-    // ANCHOR_END: here
+    let config = parse_config(&args);
 
     println!("Searching for {}", config.query);
     println!("In file {}", config.filename);
@@ -20,7 +13,11 @@ fn main() {
     let contents = fs::read_to_string(config.filename)
         .expect("Something went wrong reading the file");
 
+    // --snip--
+    // ANCHOR_END: here
+
     println!("With text:\n{}", contents);
+    // ANCHOR: here
 }
 
 struct Config {
@@ -28,15 +25,10 @@ struct Config {
     filename: String,
 }
 
-impl Config {
-    fn new(args: &[String]) -> Result<Config, &str> {
-        if args.len() < 3 {
-            return Err("not enough arguments");
-        }
+fn parse_config(args: &[String]) -> Config {
+    let query = args[1].clone();
+    let filename = args[2].clone();
 
-        let query = args[1].clone();
-        let filename = args[2].clone();
-
-        Ok(Config { query, filename })
-    }
+    Config { query, filename }
 }
+// ANCHOR_END: here
