@@ -7,7 +7,7 @@ pub struct Config {
 }
 
 impl Config {
-    pub fn new(args: &[String]) -> Result<Config, &str> {
+    pub fn new(args: &[String]) -> Result<Config, &'static str> {
         if args.len() < 3 {
             return Err("not enough arguments");
         }
@@ -26,19 +26,9 @@ pub fn run(config: Config) -> Result<(), Box<dyn Error>> {
 }
 
 // ANCHOR: here
-// ANCHOR: ch13
 pub fn search<'a>(query: &str, contents: &'a str) -> Vec<&'a str> {
-    let mut results = Vec::new();
-
-    for line in contents.lines() {
-        if line.contains(query) {
-            results.push(line);
-        }
-    }
-
-    results
+    vec![]
 }
-// ANCHOR_END: ch13
 // ANCHOR_END: here
 
 #[cfg(test)]
