@@ -1,19 +1,15 @@
-// edition:2018
-// aux-build:edition-kw-macro-2015.rs
-
-#![feature(async_closure)]
-
-fn main() {}
+// edition:2015
+// aux-build:edition-kw-macro-2018.rs
 
 #[macro_use]
-extern crate edition_kw_macro_2015;
+extern crate edition_kw_macro_2018;
 
 mod module {
-    pub fn r#async() {}
+    pub fn async() {}
 }
 
 pub fn check_async() {
-    let mut async = 1; //~ ERROR expected identifier, found keyword `async`
+    let mut async = 1; // OK
     let mut r#async = 1; // OK
 
     r#async = consumes_async!(async); // OK
@@ -21,10 +17,10 @@ pub fn check_async() {
     r#async = consumes_async_raw!(async); //~ ERROR no rules expected the token `async`
     r#async = consumes_async_raw!(r#async); // OK
 
-    if passes_ident!(async) == 1 {}
+    if passes_ident!(async) == 1 {} // OK
     if passes_ident!(r#async) == 1 {} // OK
-    module::async(); //~ ERROR expected identifier, found keyword `async`
+    module::async(); // OK
     module::r#async(); // OK
-
-    let _recovery_witness: () = 0; //~ ERROR mismatched types
 }
+
+fn main() {}
