@@ -1,11 +1,10 @@
 #![feature(if_let_guard)]
-#![allow(incomplete_features)]
 
 #[deny(irrefutable_let_patterns)]
 fn irrefutable_let_guard() {
     match Some(()) {
         Some(x) if let () = x => {}
-        //~^ ERROR irrefutable if-let guard
+        //~^ ERROR irrefutable `if let` guard
         _ => {}
     }
 }
